/*
 * Copyright (c) 2008-2016 Haulmont.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.haulmont.cuba.core.sys.persistence;

import com.haulmont.chile.core.model.MetaClass;
import com.haulmont.chile.core.model.MetaProperty;
import com.haulmont.cuba.core.Persistence;
import com.haulmont.cuba.core.entity.BaseGenericIdEntity;
import com.haulmont.cuba.core.global.Metadata;
import com.haulmont.cuba.core.global.PersistenceHelper;
import com.haulmont.cuba.security.app.EntityAttributeChanges;
import org.eclipse.persistence.jpa.JpaCache;
import org.springframework.stereotype.Component;

import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.persistence.EntityManagerFactory;

@Component(OrmCacheSupport.NAME)
public class OrmCacheSupport {

    public static final String NAME = "cuba_OrmCacheSupport";

    @Inject
    protected Metadata metadata;

    @Inject
    protected Persistence persistence;

    /**
     * Evicts an entity from cache if it has the given entity as an element of collection.
     *
     * @param entity    which is being updated and can potentially be an element of a collection
     * @param changes   changes in the entity. Null when creating and removing the entity.
     */
    public void evictMasterEntity(BaseGenericIdEntity entity, @Nullable EntityAttributeChanges changes) {
        MetaClass metaClass = metadata.getClassNN(entity.getClass());
        for (MetaProperty property : metaClass.getProperties()) {
            if (!property.getRange().isClass() || property.getRange().getCardinality().isMany())
                continue;
            MetaProperty inverseProp = property.getInverse();
            if (inverseProp == null || !inverseProp.getRange().getCardinality().isMany())
                continue;
            // the inverse property is a collection
            if (metadata.getTools().isCacheable(property.getRange().asClass())) {
                if (changes != null) {
                    for (String attributeName : changes.getOwnAttributes()) {
                        if (property.getName().equals(attributeName)) {
                            evictEntity(changes.getOldValue(attributeName));
                            break;
                        }
                    }
                } else {
                    Object masterEntity = entity.getValue(property.getName());
                    evictEntity(masterEntity);
                }
            }
        }
    }

    private void evictEntity(Object entity) {
        if (entity != null && !PersistenceHelper.isNew(entity)) {
            EntityManagerFactory emf = persistence.getEntityManager().getDelegate().getEntityManagerFactory();
            JpaCache cache = (JpaCache) emf.getCache();
            cache.evict(entity, true);
        }
    }
}
